#ident "$Id: lookup_multi.c,v 1.3 2004/01/29 16:01:22 raven Exp $"
/* ----------------------------------------------------------------------- *
 *   
 *  lookup_multi.c - module for Linux automount to seek multiple lookup
 *                   methods in succession
 *
 *   Copyright 1999 Transmeta Corporation - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

#include <alloca.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <malloc.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <sys/types.h>

#define MODULE_LOOKUP
#include "automount.h"

#define MODPREFIX "lookup(multi): "

struct module_info {
	int argc;
	const char *const *argv;
	struct lookup_mod *mod;
};

struct lookup_context {
	int n;
	const char **argl;
	struct module_info *m;
};

int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */

int lookup_init(const char *my_mapfmt, int argc, const char *const *argv, void **context)
{
	struct lookup_context *ctxt;
	char *map, *mapfmt;
	int i, j, an;

	if (!(*context = ctxt = malloc(sizeof(struct lookup_context))))
		goto nomem;

	memset(ctxt, 0, sizeof(struct lookup_context));

	if (argc < 1) {
		crit(MODPREFIX "No map list");
		return 1;
	}

	ctxt->n = 1;				/* Always at least one map */
	for (i = 0; i < argc; i++) {
		if (!strcmp(argv[i], "--"))	/* -- separates maps */
			ctxt->n++;
	}

	if (!(ctxt->m = malloc(ctxt->n * sizeof(struct module_info))) ||
	    !(ctxt->argl = malloc((argc + 1) * sizeof(const char *))))
		goto nomem;

	memset(ctxt->m, 0, ctxt->n * sizeof(struct module_info));

	memcpy(ctxt->argl, argv, (argc + 1) * sizeof(const char *));

	for (i = j = an = 0; ctxt->argl[an]; an++) {
		if (ctxt->m[i].argc == 0) {
			ctxt->m[i].argv = &ctxt->argl[an];
		}
		if (!strcmp(ctxt->argl[an], "--")) {
			ctxt->argl[an] = NULL;
			i++;
		} else {
			ctxt->m[i].argc++;
		}
	}

	for (i = 0; i < ctxt->n; i++) {
		if (!ctxt->m[i].argv[0]) {
			crit(MODPREFIX "missing module name");
			return 1;
		}
		map = strdup(ctxt->m[i].argv[0]);
		if (!map)
			goto nomem;

		if ((mapfmt = strchr(map, ',')))
			*(mapfmt++) = '\0';

		debug(MODPREFIX "[%d] calling open_lookup for map %s", i, map);
		if (!(ctxt->m[i].mod = open_lookup(map, MODPREFIX,
						   mapfmt ? mapfmt : my_mapfmt,
						   ctxt->m[i].argc - 1,
						   ctxt->m[i].argv + 1)))
			return 1;
		if (!ctxt->m[i].mod->lookup_one)
			debug(MODPREFIX
			      "[%d] module %s has no lookup_one symbol.",
			      i, ctxt->m[i].argv[0]);
	}

	*context = ctxt;
	return 0;

      nomem:
	crit(MODPREFIX "malloc: %m");
	return 1;
}

int lookup_ghost(const char *root, int ghost, time_t now, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	int i, ret, at_least_1 = 0;

	for (i = 0; i < ctxt->n; i++) {
		ret = ctxt->m[i].mod->lookup_ghost(root, ghost, now,
						   ctxt->m[i].mod->context);
		if (ret & LKP_FAIL)
			continue;

		at_least_1 = 1;	
	}

	if (!at_least_1)
		return LKP_FAIL;

	return LKP_INDIRECT;
}

int lookup_mount(const char *root, const char *name, int name_len, void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	struct mapent_cache *me;
	char *old_mapent = NULL;
	int i, ret = 0;

	/*
	 * Because the called lookup modules do not know they are being
	 * called in the context of a multi-map, we need to take special
	 * care here to make sure that a real lookup is done for the key.
	 * The lookup_file.c:lookup_mount code, for example, will not
	 * perform a real lookup if the file map hasn't changed.  However,
	 * if it is listed second in a list of maps, and there was a duplicate
	 * key, where the first instance of that key was removed, then the
	 * file map didn't change, but we need to re-read the key from
	 * the file map anyway.
	 *
	 * So, we go ahead and store the old value of the key, delete the
	 * key from the cache, and then force the modules to perform real
	 * lookups on the key (until we find one).  Then, after the mount,
	 * if the map entry in the cache has changed, we need to trigger
	 * a re-read of the map in the parent.
	 */
	me = cache_lookup_distinct(name);
	if (me) {
		/* squirrel away the old map entry */
		old_mapent = alloca(strlen(me->mapent));
		strcpy(old_mapent, me->mapent);

		ret = cache_delete(root, name, 0);
		/*
		 * cache_delete fails for the following reasons:
		 * - key doesn't exist in the cache
		 * - malloc failed
		 * - path already mounted
		 * It shouldn't be the first as we were able to successfully
		 * look up the key in order to delete it. If it's the third,
		 * then we have a bug elsewhere (as we shouldn't trigger a
		 * lookup for a path that's already mounted).  A malloc
		 * failure will likely result in other operations failing.
		 * Thus, if we fail to delete the key from the cache, we may
		 * as well give up.
		 */
		if (ret != CHE_OK) {
			info(MODPREFIX
			     "Lookup failed -- cache_delete %s %s 0 returned "
			     "error %d", root, name, ret);
			return LKP_FAIL;
		}
	}

	/* trigger the mount */
	for (i = 0; i < ctxt->n; i++) {
		struct lookup_mod *mod = ctxt->m[i].mod;
		lookup_mount_t lookup_mount = mod->lookup_mount;
		lookup_one_t lookup_one = mod->lookup_one;

		/*
		 * We removed the key from the cache, so the logic in the
		 * lookup_mount routine may see that the map hasn't changed
		 * and thus assume that the key no longer exists.  So, we
		 * force the lookup by calling lookup_one, and then call
		 * the lookup_mount routine.
		 */
		if (lookup_one)
			(void)lookup_one(root, name, name_len, mod->context);
		ret = lookup_mount(root, name, name_len, mod->context);
		if (ret == 0)
			break;
	}

	/*
	 * If the map entry changed, re-read.  Note that we use
	 * cache_lookup and not cache_lookup_distinct here.  This is
	 * to catch the case where we had an exact match before, but
	 * now we matched wild-card (or vice versa).
	 */
	me = cache_lookup(name);

	/* Sanity check: if there is no entry in the cache, the mount should
	 * not succeed. */
	if (!me && ret == 0)
		warn(MODPREFIX
		     "Error: %s does not exist in the cache, but the "
		     "mount succeeded!", name);

	/* If we had a key before, check to see if it changed or was deleted */
	if (old_mapent)
		if (!me || (me && strcmp(me->mapent, old_mapent)))
			kill(getppid(), SIGHUP);

	return ret;
}

int lookup_done(void *context)
{
	struct lookup_context *ctxt = (struct lookup_context *) context;
	int i, rv = 0;

	for (i = 0; i < ctxt->n; i++) {
		rv = rv || close_lookup(ctxt->m[i].mod);
	}

	free(ctxt->argl);
	free(ctxt->m);
	free(ctxt);

	return rv;
}
